@@ -5,6 +5,7 @@ import pathlib
 import json
 from datetime import datetime
 from django.db import models
+from django.db.models import Q
 from django.conf import settings
 
 from django.contrib.postgres.search import SearchVector
@@ -28,13 +29,22 @@ class Collection(models.Model):
     def object_data_name(self):
         return self.object_name + "_data"
 
-    def clean_documents(self, config):
+    def clean_documents(self, config, document_list=None):
         # return a reduced list of documents based on permissions defined in config
         # (WikiConfig instance)
+        #
+        # Also reduced by the documents in document_list
+
+        document_limit = Q()
+        if document_list is not None:
+            document_limit = Q(document__in=document_list)
+
         new_documents = {}
         for value in (
             Value.objects.filter(
-                document__in=config.valid_ids.all(), field__in=config.valid_fields.all()
+                document_limit,
+                document__in=config.valid_ids.all(),
+                field__in=config.valid_fields.all(),
             )
             .prefetch_related("field")
             .prefetch_related("document")
@@ -202,11 +212,17 @@ class WikiConfig(models.Model):
         for document_dict in self.collection.clean_documents(self):
             document = document_dict["db_document"]
             filtered_data = {}
-            for filter in settings.TORQUE_FILTERS:
+            for filter in getattr(settings, "TORQUE_FILTERS", []):
                 filtered_data[filter.name()] = filter.document_value(
                     document_dict["fields"]
                 )
 
+            explore_rank = None
+            if getattr(settings, "TORQUE_EXPLORE_RANK", None):
+                explore_rank = settings.TORQUE_EXPLORE_RANK.get_rank_key(
+                    document_dict["fields"]
+                )
+
             sc_documents.append(
                 SearchCacheDocument(
                     document=document,
@@ -216,6 +232,7 @@ class WikiConfig(models.Model):
                     wiki_config=self,
                     filtered_data=filtered_data,
                     data=" ".join(list(map(str, document_dict["fields"].values()))),
+                    explore_rank=explore_rank,
                 )
             )
 
@@ -397,7 +414,7 @@ class TableOfContents(models.Model):
     )
     raw = models.BooleanField(default=False)
 
-    def render_to_mwiki(self, wiki_config):
+    def render_to_mwiki(self, wiki_config, results_limit=[]):
         data = json.loads(self.json_file)
 
         global_collection_data_name = wiki_config.collection.object_data_name()
@@ -414,7 +431,13 @@ class TableOfContents(models.Model):
 
         def add_to_data(wiki_config):
             collection = wiki_config.collection
-            documents = collection.clean_documents(wiki_config)
+
+            document_limit = None
+            if results_limit:
+                key_list = [rl[1] for rl in results_limit if rl[0] == collection.name]
+                document_limit = Document.objects.filter(key__in=key_list)
+            documents = collection.clean_documents(wiki_config, document_limit)
+
             data[global_collection_data_name][collection.name] = {
                 document["fields"][collection.key_field]: document["fields"]
                 for document in documents
@@ -493,6 +516,7 @@ class SearchCacheDocument(models.Model):
     filtered_data = models.JSONField(null=True)
     data_vector = SearchVectorField(null=True)
     dirty = models.BooleanField(default=False)
+    explore_rank = models.TextField(null=True)
 
     class Meta:
         indexes = (GinIndex(fields=["data_vector"]),)
