@@ -55,44 +55,74 @@ def get_system_information(request, fmt):
         raise Exception(f"Invalid format {fmt}")
 
 
-def search(q, filters, offset, template_config, wiki_configs, fmt, multi):
+def search_data(qs, filters, wiki_configs, documents_limited_to):
+    data_vector_q = Q()
+    if qs:
+        for q in qs:
+            data_vector_q |= Q(data_vector=q)
+
+    document_limit_or = Q()
+    for (collection, document) in documents_limited_to:
+        document_limit_or |= Q(collection__name=collection, document__key=document)
+
     results = (
         models.SearchCacheDocument.objects.filter(
+            data_vector_q,
+            document_limit_or,
             collection__in=wiki_configs.values_list("collection", flat=True),
             wiki__wiki_key__in=wiki_configs.values_list("wiki__wiki_key", flat=True),
             group__in=wiki_configs.values_list("group", flat=True),
             wiki_config__in=wiki_configs,
-            data_vector=q,
         )
         .select_related("document")
         .select_related("collection")
     )
     filter_results = []
-    for filter in settings.TORQUE_FILTERS:
+    torque_filters_by_name = {
+        f.name(): f for f in getattr(settings, "TORQUE_FILTERS", [])
+    }
+    for filter in getattr(settings, "TORQUE_FILTERS", []):
         additional_filters = []
         for name, values in filters.items():
             if name != filter.name():
                 q_objects = Q()
                 for value in values:
-                    q_dict = {"filtered_data__%s" % name: value}
+                    if torque_filters_by_name[name].is_list():
+                        key = "filtered_data__%s__contains" % name
+                    else:
+                        key = "filtered_data__%s" % name
+                    q_dict = {key: value}
                     q_objects |= Q(**q_dict)
                 additional_filters.append(q_objects)
         filter_result = {
             "name": filter.name(),
             "display": filter.display_name(),
+            "primary": filter.is_primary(),
             "counts": {},
         }
+
         grouped_results = (
             results.filter(*additional_filters)
             .values("filtered_data__%s" % filter.name())
             .annotate(total=Count("id"))
         )
         for result in grouped_results:
-            name = result["filtered_data__%s" % filter.name()] or "None"
-            filter_result["counts"][name] = {
-                "name": name,
-                "total": result["total"],
-            }
+            name = result["filtered_data__%s" % filter.name()]
+            if name:
+                if filter.is_list():
+                    for item in name:
+                        if item not in filter_result["counts"]:
+                            filter_result["counts"][item] = {
+                                "name": item,
+                                "total": 0,
+                            }
+
+                        filter_result["counts"][item]["total"] += result["total"]
+                else:
+                    filter_result["counts"][name] = {
+                        "name": name,
+                        "total": result["total"],
+                    }
 
         names = list(filter_result["counts"].keys())
         names = filter.sort(names)
@@ -109,17 +139,38 @@ def search(q, filters, offset, template_config, wiki_configs, fmt, multi):
     additional_filters = []
     for name, values in filters.items():
         q_objects = Q()
+        if torque_filters_by_name[name].is_list():
+            key = "filtered_data__%s__contains" % name
+        else:
+            key = "filtered_data__%s" % name
         for value in values:
-            q_dict = {"filtered_data__%s" % name: value}
+            q_dict = {key: value}
             q_objects |= Q(**q_dict)
         additional_filters.append(q_objects)
 
-    returned_results = (
-        results.filter(*additional_filters)
-        .annotate(rank=SearchRank(F("data_vector"), SearchQuery(q)))
-        .order_by("-rank")
-    )
+    if qs:
+        returned_results = (
+            results.filter(*additional_filters)
+            .annotate(rank=SearchRank(F("data_vector"), SearchQuery(q[0])))
+            .order_by("-rank")
+        )
+    elif getattr(settings, "TORQUE_EXPLORE_RANK", None):
+        returned_results = results.filter(*additional_filters).order_by("explore_rank")
+    else:
+        returned_results = results.filter(*additional_filters)
 
+    return (returned_results, filter_results)
+
+
+def search(
+    q, filters, offset, template_config, wiki_configs, documents_limited_to, fmt
+):
+    # search_data requires filter values to be lists
+    filters = {name: [value] for (name, value) in filters.items()}
+
+    (returned_results, filter_results) = search_data(
+        [q], filters, wiki_configs, documents_limited_to
+    )
     # While this result isn't actually mwiki text, this result is intended
     # for the mediawiki Torque extension.  Probably better to keep
     # the mwiki format than to do something like create a new "torque" format.
@@ -171,6 +222,8 @@ def search(q, filters, offset, template_config, wiki_configs, fmt, multi):
 def search_global(request, fmt):
     q = request.GET["q"]
     f = json.loads(request.GET["f"]) if "f" in request.GET and request.GET["f"] else {}
+    results_limit_json = json.loads(request.body) if request.body else None
+    results_limit = results_limit_json if results_limit_json else []
     offset = int(request.GET.get("offset", 0))
     group = request.GET["group"]
     global_wiki_key = request.GET["wiki_key"]
@@ -185,19 +238,129 @@ def search_global(request, fmt):
     configs = models.WikiConfig.objects.filter(
         collection__name__in=collection_names, wiki__wiki_key__in=wiki_keys, group=group
     ).all()
-    return search(q, f, offset, global_config, configs, fmt, True)
+    return search(q, f, offset, global_config, configs, results_limit, fmt)
 
 
 def search_collection(request, collection_name, fmt):
     q = request.GET["q"]
     f = json.loads(request.GET["f"]) if "f" in request.GET and request.GET["f"] else {}
+    results_limit_json = json.loads(request.body) if request.body else None
+    results_limit = (
+        [[collection_name, key] for key in results_limit_json]
+        if results_limit_json
+        else []
+    )
     offset = int(request.GET.get("offset", 0))
     group = request.GET["group"]
     wiki = get_wiki_from_request(request, collection_name)
     configs = models.WikiConfig.objects.filter(
         collection__name=collection_name, wiki=wiki, group=group
     )
-    return search(q, f, offset, configs.first(), configs, fmt, False)
+    return search(q, f, offset, configs.first(), configs, results_limit, fmt)
+
+
+def explore(
+    qs, filters, offset, template_name, template_config, wiki_configs, results_limit
+):
+    (returned_results, filter_results) = search_data(
+        qs, filters, wiki_configs, results_limit
+    )
+
+    mwiki_text = ""
+
+    result_set = [
+        r.document.to_dict(r.wiki_config)
+        for r in returned_results[offset : (offset + 100)]
+    ]
+    explore_templates = models.Template.objects.filter(
+        type="Explore",
+        collection=template_config.collection,
+    )
+    if explore_templates:
+        if template_name:
+            template = jinja_env.from_string(
+                explore_templates.get(name=template_name).get_file_contents()
+            )
+        else:
+            template = jinja_env.from_string(
+                explore_templates.get(is_default=True).get_file_contents()
+            )
+        mwiki_text += template.render(
+            {template_config.collection.object_data_name(): result_set}
+        )
+    else:
+        for result in result_set:
+            template = jinja_env.from_string(
+                models.Template.objects.get(
+                    name="Search", collection=template_config.collection
+                ).get_file_contents()
+            )
+            mwiki_text += template.render(
+                {template_config.collection.object_name: result}
+            )
+            mwiki_text += "\n\n"
+
+    return JsonResponse(
+        {
+            "num_results": returned_results.count(),
+            "results": [(r.collection.name, r.document.key) for r in returned_results],
+            "mwiki_text": mwiki_text,
+            "filter_results": filter_results,
+        },
+        safe=False,
+    )
+
+
+def explore_global(request):
+    qs = (
+        json.loads(request.GET["qs"])
+        if "qs" in request.GET and request.GET["qs"]
+        else []
+    )
+    f = json.loads(request.GET["f"]) if "f" in request.GET and request.GET["f"] else {}
+    template_name = request.GET["template"]
+    results_limit_json = json.loads(request.body) if request.body else None
+    results_limit = results_limit_json if results_limit_json else []
+    offset = int(request.GET.get("offset", 0))
+    group = request.GET["group"]
+    global_wiki_key = request.GET["wiki_key"]
+    global_collection_name = request.GET["collection_name"]
+    wiki_keys = request.GET["wiki_keys"].split(",")
+    collection_names = request.GET["collection_names"].split(",")
+    global_config = models.WikiConfig.objects.get(
+        collection__name=global_collection_name,
+        wiki__wiki_key=global_wiki_key,
+        group=group,
+    )
+    configs = models.WikiConfig.objects.filter(
+        collection__name__in=collection_names, wiki__wiki_key__in=wiki_keys, group=group
+    ).all()
+    return explore(qs, f, offset, template_name, global_config, configs, results_limit)
+
+
+def explore_collection(request, collection_name):
+    qs = (
+        json.loads(request.GET["qs"])
+        if "qs" in request.GET and request.GET["qs"]
+        else []
+    )
+    results_limit_json = json.loads(request.body) if request.body else None
+    results_limit = (
+        [[collection_name, key] for key in results_limit_json]
+        if results_limit_json
+        else []
+    )
+    f = json.loads(request.GET["f"]) if "f" in request.GET and request.GET["f"] else {}
+    template_name = request.GET["template"]
+    offset = int(request.GET.get("offset", 0))
+    group = request.GET["group"]
+    wiki = get_wiki_from_request(request, collection_name)
+    configs = models.WikiConfig.objects.filter(
+        collection__name=collection_name, wiki=wiki, group=group
+    )
+    return explore(
+        qs, f, offset, template_name, configs.first(), configs, results_limit
+    )
 
 
 def edit_record(collection_name, key, group, wiki, field, new_value):
@@ -305,6 +468,10 @@ def get_collection(request, collection_name, fmt):
 
 def get_toc(request, collection_name, toc_name, fmt):
     group = request.GET["group"]
+
+    results_limit_json = json.loads(request.body) if request.body else None
+    results_limit = results_limit_json if results_limit_json else []
+
     wiki = get_wiki_from_request(request, collection_name)
     collection = models.Collection.objects.get(name=collection_name)
 
@@ -323,11 +490,13 @@ def get_toc(request, collection_name, toc_name, fmt):
         return HttpResponse(status=403)
 
     if fmt == "mwiki":
-        return HttpResponse(toc.render_to_mwiki(wiki_config))
+        return HttpResponse(toc.render_to_mwiki(wiki_config, results_limit))
     elif fmt == "html":
-        cached_toc = wiki_config.cached_tocs.get(toc=toc)
-
-        return HttpResponse(cached_toc.rendered_html)
+        if not results_limit:
+            cached_toc = wiki_config.cached_tocs.get(toc=toc)
+            return HttpResponse(cached_toc.rendered_html)
+        else:
+            return HttpResponse()
     else:
         raise Exception(f"Invalid format {fmt}")
 
@@ -373,7 +542,7 @@ def get_document(group, wiki, key, fmt, collection_name, view=None):
         wiki=wiki,
         group=group,
     )
-    document = models.Document.objects.get(key=key, collection=collection).to_dict(
+    document = wiki_config.valid_ids.get(key=key, collection=collection).to_dict(
         wiki_config
     )
 
@@ -746,11 +915,11 @@ def add_csv(request):
     name = determine_name()
     post_fields = json.loads(request.body)
 
-    documents = []
+    massive_or = Q()
     for post_doc in post_fields["documents"]:
-        documents.append(
-            models.Document.objects.get(collection__name=post_doc[0], key=post_doc[1])
-        )
+        massive_or |= Q(collection__name=post_doc[0], key=post_doc[1])
+
+    documents = models.Document.objects.filter(massive_or)
 
     csv_spec = models.CsvSpecification(
         name=name, filename=post_fields["filename"], fields=post_fields["fields"]
@@ -771,17 +940,13 @@ def get_csv(request, name, fmt):
 
     group = request.GET["group"]
 
-    documents = csv_spec.documents.all()
-    valid_documents = []
-    wiki_configs_for_csv = set()
-    for document in documents:
-        if document.wiki_config.filter(group=group).count() > 0:
-            valid_documents.append(document)
-            wiki_configs_for_csv.add(document.wiki_config.get(group=group))
+    valid_documents = csv_spec.documents.filter(
+        wiki_config__group=group
+    ).prefetch_related("collection")
 
     if fmt == "json":
         document_information = {}
-        for document in documents:
+        for document in valid_documents:
             if document.collection.name not in document_information:
                 document_information[document.collection.name] = []
             document_information[document.collection.name].append(document.key)
@@ -794,15 +959,20 @@ def get_csv(request, name, fmt):
             }
         )
     elif fmt == "csv":
-        field_names = sorted(csv_spec.fields)
+        csv_field_names = sorted(csv_spec.fields)
 
-        valid_field_names = []
-        for wiki_config in wiki_configs_for_csv:
-            for field_name in field_names:
-                if wiki_config.valid_fields.filter(name=field_name).count() > 0:
-                    valid_field_names.append(field_name)
+        wiki_configs_for_csv = models.WikiConfig.objects.filter(
+            group=group, valid_ids__in=csv_spec.documents.all()
+        ).distinct()
 
-        valid_field_names = sorted(list(set(valid_field_names)))
+        valid_field_names = [
+            field.name
+            for field in models.Field.objects.filter(
+                wiki_config__in=wiki_configs_for_csv
+            ).distinct()
+        ]
+
+        field_names = [fn for fn in csv_field_names if fn in valid_field_names]
 
         response = HttpResponse(
             content_type="text/csv",
@@ -814,7 +984,7 @@ def get_csv(request, name, fmt):
         writer = csv.writer(response)
 
         columns = []
-        for field_name in valid_field_names:
+        for field_name in field_names:
             if field_name in getattr(settings, "TORQUE_CSV_PROCESSORS", {}):
                 columns.extend(
                     settings.TORQUE_CSV_PROCESSORS[field_name].field_names(field_name)
@@ -823,19 +993,28 @@ def get_csv(request, name, fmt):
                 columns.append(field_name)
         writer.writerow(columns)
 
+        all_pertinent_values = (
+            models.Value.objects.filter(
+                field__name__in=field_names, document__in=valid_documents
+            )
+            .prefetch_related("field")
+            .prefetch_related("document")
+        )
+
+        values_by_documents_and_fields = {}
+
+        for value in all_pertinent_values:
+            if value.document not in values_by_documents_and_fields:
+                values_by_documents_and_fields[value.document] = {}
+            values_by_documents_and_fields[value.document][value.field.name] = value
+
         for document in valid_documents:
             row = []
-            values_by_field = {
-                v.field.name: v
-                for v in document.values.filter(
-                    field__name__in=valid_field_names
-                ).prefetch_related("field")
-            }
-            for field_name in valid_field_names:
+            values_by_field = values_by_documents_and_fields[document]
+            for field_name in field_names:
                 python_value = None
                 if field_name in values_by_field:
                     python_value = values_by_field[field_name].to_python()
-
                 if python_value and field_name in getattr(
                     settings, "TORQUE_CSV_PROCESSORS", {}
                 ):
